struct assocarray {
	char			*name;
	unsigned char		*macro;
	struct assocarray	*prev;
	struct assocarray	*next;
};
typedef struct assocarray Aarray;
typedef struct assocarray *AarrayPtr;
typedef struct assocarray **AarrayPPtr;

#define		NULLASSOC		((AarrayPtr)0)

AarrayPtr assoc_find (AarrayPtr pool, char *label);
unsigned char * assoc_lookup (AarrayPtr pool, char *label);
void assoc_list (AarrayPtr pool, int maxlen, char *label);
int assoc_addstr (AarrayPPtr pool, char *label, char *value, int sorted);
int assoc_free (AarrayPPtr pool, char *label);
